/***
 * Excerpted from "Modern Systems Programming with Scala Native",
 * published by The Pragmatic Bookshelf.
 * Copyrights apply to this code. It may not be used to create training material,
 * courses, books, articles, and the like. Contact us if you are in doubt.
 * We make no guarantees that this code is fit for any purpose.
 * Visit http://www.pragmaticprogrammer.com/titles/rwscala for more book information.
***/
import scala.scalanative.unsafe._
import scala.scalanative.libc._

// import scala.scalanative.native._
import stdio._, string._, stdlib._
import scala.scalanative.posix.unistd._
import scala.scalanative.posix.sys.socket._
import scala.scalanative.posix.netinet.in._
import scala.scalanative.posix.arpa.inet._
import scala.collection.mutable
import scala.scalanative.posix.netdb._
import scala.scalanative.posix.netdbOps._

object main {
  def makeConnection(address:CString, port:CString):Int = {
    val hints = stackalloc[addrinfo]
    string.memset(hints.asInstanceOf[Ptr[Byte]], 0, sizeof[addrinfo])
    hints.ai_family = AF_UNSPEC
    hints.ai_socktype = SOCK_STREAM

    val addrInfo_ptr:Ptr[Ptr[addrinfo]] = stackalloc[Ptr[addrinfo]]
    println("about to perform lookup")
    val lookup_result = getaddrinfo(address, port, hints, addrInfo_ptr)
    println(s"lookup returned ${lookup_result}")
    if (lookup_result != 0) {
      val errString = util.gai_strerror(lookup_result)
      stdio.printf(c"errno: %d %s\n", lookup_result, errString)
      throw new Exception("no address found")
    } else {
      val addrInfo = !addrInfo_ptr
      stdio.printf(c"""got addrinfo: flags %d, family %d, socktype %d,
      protocol %d\n""", addrInfo.ai_family, addrInfo.ai_flags,
      addrInfo.ai_socktype, addrInfo.ai_protocol)

      println("creating socket")
      val sock = socket(addrInfo.ai_family, addrInfo.ai_socktype,
                        addrInfo.ai_protocol)
      println(s"socket returned fd $sock")
      if (sock < 0) {
        throw new Exception("error in creating socket")
      }

      println("connecting")
      val connect_result = connect(sock, addrInfo.ai_addr, addrInfo.ai_addrlen)
      println(s"connect returned $connect_result")
      if (connect_result != 0) {
        val err = errno.errno
        val errString = string.strerror(err)
        stdio.printf(c"errno: %d %s\n", err, errString)

        throw new Exception("connection failed")
      }
      sock
    }
  }

  def makeRequest(sock:Ptr[FILE],request:String):String = {
    val resp_buffer = stdlib.malloc(2048)
    val resp = Zone { implicit z =>
      val request_cstring = toCString(request)
      stdio.fprintf(sock, request_cstring)
      val resp_cstring = fgets(resp_buffer, 4095, sock)
      fromCString(resp_buffer)
    }
    stdlib.free(resp_buffer)
    resp
  }

  def handleConnection(sock:Int): Unit = {
    val resp_buffer = malloc(4096)
    val socket_fd = util.fdopen(sock, c"r+")
    val resp = makeRequest(socket_fd,"hello?  is anybody there?\n")
    println(s"I got a response: ${resp.trim()}")
    fclose(socket_fd)
    println("done")
  }

  def main(args: Array[String]):Unit = {
    if (args.length != 2) {
      println("Usage: ./tcp_test [address] [port]")
      return ()
    }

    val sock = Zone { implicit z =>
      val address = toCString(args(0))
      val port = toCString(args(1))

      stdio.printf(c"looking up address: %s port: %s\n", address, port)

      makeConnection(address, port)
    }
    handleConnection(sock)
  }
}

@extern
object util {
  def getaddrinfo(address:CString, port:CString, hints: Ptr[addrinfo], res: Ptr[Ptr[addrinfo]]):Int = extern
  def socket(family:Int, socktype:Int, protocol:Int):Int = extern
  def connect(sock:Int, addrInfo:Ptr[sockaddr], addrLen:CInt):Int = extern
  def gai_strerror(code:Int):CString = extern
  def fdopen(fd:Int, mode:CString):Ptr[FILE] = extern
}
