/***
 * Excerpted from "Modern Systems Programming with Scala Native",
 * published by The Pragmatic Bookshelf.
 * Copyrights apply to this code. It may not be used to create training material,
 * courses, books, articles, and the like. Contact us if you are in doubt.
 * We make no guarantees that this code is fit for any purpose.
 * Visit http://www.pragmaticprogrammer.com/titles/rwscala for more book information.
***/
object curlMulti {
  // STARt:onPoll
  def onPoll(pollHandle:PollHandle, status:Int, events:Int):Unit = {
    val socket = !(pollHandle.cast[Ptr[Ptr[Byte]]])
    val actions = (events & 1) | (events & 2) // Whoa, nelly!
    val running_handles = stackalloc[Int]
    val result = multi_socket_action(multi, socket, actions, running_handles) // check
  }

  def on_timer(handle:TimerHandle):Unit = {
    val running_handles = stackalloc[Int]
    multi_socket_action(multi,-1.cast[Ptr[Byte]],0,running_handles)
    logger.debug(s"on_timer fired, ${!running_handles} sockets running")
  }

  val writeCB = CFunctionPtr.fromFunction4(writeData)
  val headerCB = CFunctionPtr.fromFunction4(writeHeader)
  val readyCB = CFunctionPtr.fromFunction3(ready_for_curl)
  val timerCB = CFunctionPtr.fromFunction1(on_timer)

  def socket_state_change(curl:Curl, socket:Ptr[Byte], action:Int, data:Ptr[Byte], socket_data:Ptr[Byte]):Int = {
    logger.debug(s"handle_socket called with action ${action}")
    // logger.trace(c"\tdata %p, socket %p, action %d\n", socket, socket_data, action)
    val pollHandle = if (socket_data == null) {
      logger.debug("uninitialized poll_handle")
      val buf = malloc(uv_handle_size(UV_POLL_T)).cast[Ptr[Ptr[Byte]]]
      !buf = socket
      check(uv_poll_init_socket(loop, buf, socket), "uv_poll_init_socket")
      debug_check("multi_assign",multi_assign(multi, socket, buf.cast[Ptr[Byte]]))
      // printf(c"\t assigned data %x containing socket %x to socket %x\n", buf, !buf, socket)
      buf
    } else {
      logger.trace("poll_handle exists")
      socket_data.cast[Ptr[Ptr[Byte]]]
    }

    val events = action match {
      case POLL_NONE => None
      case POLL_IN => Some(UV_READABLE)
      case POLL_OUT => Some(UV_WRITABLE)
      case POLL_INOUT => Some(UV_READABLE | UV_WRITABLE)
      case POLL_REMOVE => None
    }

    events match {
      case Some(ev) =>
        logger.info(s"starting poll with events $ev")
        uv_poll_start(pollHandle, ev, readyCB)
      case None =>
        logger.warn("stopping poll")
        uv_poll_stop(pollHandle)
        start_timer(multi, 1, null)
        // uv_timer_start(timer_handle, timerCB, 1, 0)
        // free(pollHandle)
    }
    0
  }

  def cleanup_requests():Unit = {
    val messages = stackalloc[Int]
    val privateDataPtr = stackalloc[Ptr[CurlRequest]]
    var message:Ptr[CurlMessage] = multi_info_read(multi,messages)
    while (message != null) {
      logger.info(s"Got a message ${!message._1} from multi_info_read, ${!messages} left in queue")
      val handle = !message._2
      trace_check("easy_getinfo",easy_getinfo(handle, GET_PRIVATEDATA, privateDataPtr))
      // val privateData = !privateDataPtr
      Curl.complete_request(!privateDataPtr)
      message = multi_info_read(multi,messages)
    }
    logger.debug("done handling messages")
  }

  def shutdown():Unit = {
    debug_check("cleanup",multi_cleanup(multi))
    global_cleanup()
    initialized = false
  }

  def set_timeout(curl:MultiCurl, timeout_ms:Long, data:Ptr[Byte]):Int = {
    logger.info(s"start_timer called with timeout ${timeout_ms} ms")
    val time = if (timeout_ms < 1) {
      logger.warn("setting effective timeout to 1")
      1
    } else timeout_ms
    // printf(c"timer handle is %p, timer cb is %p\n", timer_handle, timerCB)
    check(uv_timer_start(timer_handle, timerCB, time, 0), "uv_timer_start")
    cleanup_requests()
    0
  }

  val socketCB = CFunctionPtr.fromFunction5(handle_socket)
  val curltimerCB = CFunctionPtr.fromFunction3(start_timer)

  def init():Unit = {
    if (initialized == false) {
      loop = uv_default_loop()
      global_init(1)
      multi = multi_init()
      val setopt_r_1 = multi_setopt(multi, SOCKETFUNCTION, socketCB)
      logger.debug(s"multi_setopt SOCKETFUNCTION ${setopt_r_1}")
      val setopt_r_2 = multi_setopt(multi, TIMERFUNCTION, curltimerCB)
      logger.debug(s"multi_setopt TIMERFUNCTION ${setopt_r_2}")

      timer_handle = malloc(uv_handle_size(UV_TIMER_T))
      check(uv_timer_init(loop,timer_handle),"uv_timer_init")
      logger.debug("CURL INITIALIZED")
      initialized = true
    }
  }
}
